\ Interrupts for MSP430 target

((
Copyright (c) 2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
))

\ ============
\ *! intrpt430
\ *T MSP430 Interrupts
\ ============
\ *P The file *\i{MSP430\Intrpt430.fth} contains high level Forth
\ ** interrupt handlers which run normal Forth words. Low level
\ ** (assembler) interrupt routines can be be written using the
\ ** Forth assembler in the normal way.

\ *P In order to reduce overall RAM consumption in multi-tasking
\ ** systems, high level interrupt handlers all share a common
\ ** "stack of stacks". On entry to the interrupt handler, the
\ ** next stack is allocated. It is assumed that an interrupt
\ ** handler will not be interrupted by another interrupt at the
\ ** same level. If this is required, the return stack pointers
\ ** of each banked interrupt must be pre-allocated and set up by
\ ** an interrupt initialisation routine *\fo{INIT-INT-USER}
\ ** which is called by the start-up code. This will require
\ ** each interrupt routine to have its own stack area, and for
\ ** the user areas to be allocated dynamically.

\ *P The number of stacks allocated is defined by the equate
\ ** *\fo{#IRQs} in the control file.

\ *P Note that each interrupt has its own *\fo{USER} area, but
\ ** that *\b{no user variables are initialised}.


\ ***********************
\ *S Interrupt management
\ ***********************

hex

[undefined] ei [if]
code ei         \ -- ; enable interrupts
\ *G Global enable interrupts.
  eint    ret
end-code
[then]

[undefined] di [if]
code di         \ -- ; disable interrupts
\ *G Global disable interrupts.
  dint   ret
end-code
[then]

(( for DOCGEN
code save-int   \ -- n ; save interrupt status
\ *G Geturn interrupt status and then disable interrupts.
\ ** Use *\fo{[I} and *\fo{I]} for new code.
\ ** Moved to *\i{LIB430.FTH}.
  sub    # 2  psp
  mov    tos  0 (psp)
  mov    sr  tos
  dint
  ret
end-code

code restore-int        \ n -- ; restore interrupt
\ *G Restore state returned by *\fo{SAVE-INT}.
\ ** Use *\fo{[I} and *\fo{I]} for new code.
\ ** Moved to *\i{LIB430.FTH}.
  mov    tos  sr
  mov    @psp+  tos
  ret
end-code
))

[undefined] [I [if]
code [I		\ R: -- x ; preserve I/F status
\ *G Preserve status on return stack, disable interrupts.
\ ** The state is restored by *\fo{I]}.
  pop    r14
  push   sr
  dint
  br     r14
end-code

code I]		\ R: x -- ; restore int. status from r. stack
\ *G Restore status saved by *\fo{[I} from the return stack.
  pop    r14
  pop    sr
  br     r14
end-code
[then]


\ *P On the MSP430 the following register usage is the default:
\ *E Forth        MSP430        Comments
\ ** -----------------------------------
\ ** IP           R0/PC         MSP430 PC
\ ** RSP          R1/SP         MSP430 return stack
\ ** --           R2/CG1/SR
\ ** --           R3/CG2
\ ** PSP          R4            data stack pointer
\ ** TOS          R5            cached top of data stack
\ ** UP           R6            USER area pointer
\ ** LP (locals)  R7            points to LOCALs on r. stack
\ ** scratch      R8..R13       with R5 forms working stack
\ ** codegen      R14           code generator temporary
\ ** codegen      R15           preserves SR in shuffle
\ ** ------------------------------------

variable irq-stack-ptr
  TOP-IRQ irq-stack-ptr !

AsmCode
l: next-int	\ R15 = address of Forth word to run
\ *G Common code used by all Forth high-level interrupt handlers.
\ save Forth state not preserved by IRQ (PC/R0 and SR/R2, SP=R1, R2/R3=CGn)
  push   r14
  push   r13
\ Switch to interrupt stacks
  mov .w & irq-stack-ptr  r13
  mov .w r13  r14
  sub .w # 2  r14
  mov .w sp  0 (r14)
\ r14 = top of new RSP, which has old SP on it
  mov .w r14  sp
  push   r12
  push   r11
  push   r10
  push   r9
  push   r8
  push   r7
  push   r6
  push   r5
  push   r4
\ make new data stack
  sub .w # IRQ-RP-size  r13
  mov .w r13  psp
\ make new user area
  sub .w # IRQ-SP-size IRQ-UP-size +  r13
  mov .w r13  up
\ update irq stack pointer for nesting
  mov .w r13  & irq-stack-ptr

\ Go run Forth word
  call   r15			\ go run Forth

\ restore from current SP stack
  pop    r4
  pop    r5
  pop    r6
  pop    r7
  pop    r8
  pop    r9
  pop    r10
  pop    r11
  pop    r12
\ switch back to original stack
  pop    r13			\ old stack pointer
  mov .w sp  & irq-stack-ptr
  mov .w r13  sp
  pop    r13
  pop    r14
  pop    r15
  reti
End-Code


\ ********************
\ *S Timer interrupt example
\ ********************
\ *P For full details of this code see *\i{TickWDT.fth}.

((
[undefined] ticks [if]
variable <ticks>

: ticks		( -- n )  <ticks> @  ;
\ *G Return current clock value in milliseconds. This value
\ ** can treated as a 32 bit unsigned value that will wrap
\ ** when it overflows.
compiler
: ticks		( -- n )  <ticks> @  ;
target
[endif]

: ticker-isr	\ --
\ *G Ticker interrupt handler
  #16 <ticks> +!
;

Proc TIMER-WDT	\ --
\ *G Assembler wrapper for the Forth handler.
   push   r15
   mov .w # ' ticker-isr  r15
   jmp    next-int
End-Code

TIMER-WDT wdt_vec !
\ *G Set the relevant interrupt vector.


: Start-Clock	\ --
\ *G Start the ticker interrupt.
  $5A16 WDTCTL !			\ WDT 16ms, ACLK, interval timer
  $01 IE1 bor!  ei			\ enable WDT interrupt
;
))

\ ***********************
\ *S Some common problems
\ ***********************
\ *P There are a few common problems that might cause an interrupt not
\ ** to work correctly:
\ *(
\ *B  a stack fault
\ *B  the source is not cleared
\ *B  the interrupts are not enabled
\ *)

\ ===============
\ *N Stack faults
\ ===============
\ *P An interrupt service routine can use the stack while it is
\ ** executing, but must clear up the stack before returning from the
\ ** interrupt.  The normal symptom of a stack fault is that the
\ ** interrupt handler runs but then the target board crashes, either
\ ** immediately or after a length of time.

\ ========================
\ *N Source is not cleared
\ ========================
\ *P Once an interrupt handler is triggered by an interrupt, the source
\ ** of the interrupt must be told that the interrupt is being
\ ** serviced.  If this is not done, the source of the interrupt will
\ ** carry on generating interrupts.  Normally this appears as the
\ ** interrupt handler executing once and then the target board
\ ** `locking'.

\ =============================
\ ** Interrupts are not enabled
\ =============================
\ *P Interrupts need to be enabled with *\fo{EI} before any interrupts will
\ ** be serviced. The vectors must be set up or the interrupt handler
\ ** assigned to the deferred word before the interrupts are enabled.


\ ======
\ *> ###
\ ======

decimal

