\ MSP430 ROM PowerForth - Multi-Tasker

((
Copyright (c) 2003, 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20070818 SFP002 Overhauled for DocGen.
		Fixed HALT and RESTART.
20020905 SFP001 First release


Target Model:
=============
Forth           MSP430          Comments
----------------------------------------
IP		R0/PC		MSP430 PC
RSP             R1/SP           MSP430 return stack
--		R2/CG1/SR
--		R3/CG2
PSP             R4
TOS             R5
UP              R6		USER area pointer
LP (locals)     R7		points to LOCAL frame on return stack
scratch		R8..R13		with R5 forms working stack
codegen		R14		code generator temporary
codegen		R15		preserves SR in shuffle
----------------------------------------
))

decimal

\ ===========
\ *! multi430
\ *T MSP430 multitasker
\ ===========
\ *P The MSP430 multitasker follows the model introduced
\ ** with the v6.1 compilers.


\ ****************************
\ *S TCB data structure layout
\ ****************************
\ *E cell       LINK    link to next task
\ ** cell       SSP     Saved Stack Pointer
\ ** cell       STAT    BIT 0   1 = running, 0 = halted
\ **                    BIT 1   1 = message pending
\ **                    BIT 2   1 = event triggered
\ **                    BIT 3   1 = event handler run
\ ** 	        others  1 = set to run task, available to user
\ ** cell       TASK    Task # that sent message here
\ ** cell       MESG    Message address
\ ** cell       EVNTw   CFA of word run as event handler
\ *P This structure is allocated at the start of the *\fo{USER}
\ ** area. Consequently the *\fo{TCB} of the current task is
\ ** given by *\fo{UP}.


\ *************
\ *S Configuration
\ *************

[undefined] test-multi? [if]
0 equ test-multi?	\ true to compile test code
\ *G If previously undefined, *\fo{TEST-MULTI?} is set to zero
\ ** and test code is not compiled.
[then]


\ *************************
\ *S Define TCB area structure
\ *************************

struct /TCB	\ -- size
\ *G The structure used by the code that matches the
\ ** description above.
  ptr tcb.link		\ link to next task ; MUST BE FIRST
  ptr tcb.ssp	        \ Saved Stack Pointer
  int tcb.status	\ status word
  int tcb.msrc		\ message source
  int tcb.mesg		\ message
  ptr tcb.event		\ xt of word which is event handler
end-struct

\ bit masks for the status cell
$0001 equ run-mask  run-mask invert equ ~run-mask	\ running
  0 equ run-bit#
$0002 equ msg-mask  msg-mask invert equ ~msg-mask	\ message available
  1 equ msg-bit#
$0004 equ trg-mask  trg-mask invert equ ~trg-mask	\ event triggered
  2 equ trg-bit#
$0008 equ evt-mask  evt-mask invert equ ~evt-mask	\ event run
  3 equ evt-bit#


\ ******************
\ Consistency checks
\ ******************

/tcb tcb-size <> [if]
  cr ." Control file and tasker disagree as to TCB size"  abort
[endif]


\ ************************
\ *S Task handling primitives
\ ************************

cdata		\ so that comma is to code space

init-u0 constant main	\ -- addr ; tcb of main task
\ *G Returns the base address of the main task's *\fo{USER} area.

variable <multi?>	\ -- addr
\ *G Returns true if the tasker is enabled.

: multi?	\ -- flag
\ *G Returns true if the tasker is enabled.
  <multi?> @
;
compiler
: multi?	( -- flag )  <multi?> @  ;
target

: single	\ -- ; disable scheduler
\ *G Disable scheduler.
  0 to multi?
;

: multi		\ -- ; enable scheduler
\ *G Enable scheduler.
  -1 to multi?
;

code pause	\ -- ; the scheduler itself
\ *G The software scheduler itself.
l: [schedule]
  cmp    # 0  & ' multi? >body		\ single or multi ?
  jeq    xschedule
\
\ save previously running task ip, tos, psp, lp, rsp
\
                                        \ save IP - already done with call here
  push   tos				\ save TOS
  push   psp				\ save PSP
  push   lp				\ save LP
  mov    rsp  0 tcb.ssp (up)		\ save SP in TCB
\
\ select next task to run
\
  begin,
    mov    0 tcb.link (up) up		\ next task
    cmp    # 0  0 tcb.status (up)	\ inspect status
  ne, until,                            \ check task is running

\
\ run selected task - rsp, lp, psp, tos, ip (PC on r.stack)
\
  mov    0 tcb.ssp (up)  rsp		\ restore SP from TCB
  pop    lp				\ restore LP
  pop    psp				\ restore PSP
  pop    tos				\ restore TOS

\
\ event handler exit
\
event-handler? [if]	\ if user wants the event handler
  bit    # trg-mask  0 tcb.status (up)	\ inspect event trigger bit in status
  ne, if,
    bic    # trg-mask 0 tcb.status (up)	\ clear event trigger bit in status
    bis    # evt-mask 0 tcb.status (up)	\ clear event trigger bit in status
    mov    0 tcb.event (up)  pc		\ run event handler
  endif,
[then]

\ main exit
l: xschedule
  ret                           	\ back to Forth (IP)
end-code

code status	\ -- task-status
\ *G Returns the current task's status cell.
  sub    # 2 psp			\ preserve TOS
  mov    tos  0 (psp)
  mov    0 tcb.status (up)  tos		\ get status
  bic    # run-mask  tos		\ clear run bit
  ret
end-code

code restart    \ task -- ; mark task TCB as running
\ *G Sets the RUN bit in the task's status cell.
  bis    # run-mask  0 tcb.status (tos)	\ mark as running
  mov    @psp+  tos
  ret
end-code

code halt       \ task -- ; reset running bit in TCB
\ *G Clears the RUN bit in the task's status cell.
  bic    # run-mask  0 tcb.status (tos)	\ mark as running
  mov    @psp+  tos
  ret
end-code

: stop          \ -- ; halt oneself
\ *G *\fo{HALT}'s the current task, and executes *\fo{PAUSE}.
  self halt  pause
;


\ **************
\ *S Event handling
\ **************
\ *P Event handling is only compiled if the equate
\ ** *\fo{EVENT-HANDLER?} is set non-zero in the control
\ ** file.

event-handler? [if]

code set-event	\ task -- ; set event trigger in task TCB
\ *G Set the event trigger in task TCB.
  bis    # trg-mask  0 tcb.status (tos)	\ mark as running
  mov    @psp+  tos
  ret
end-code

: event?	\ task -- flag ; true if task had event
\ *G Returns true if true if task has received an event
\ ** trigger which has not been cleared yet.
  tcb.status @  trg-mask and 0<>
;

((
: clr-event-run	\ -- ; reset own EVENT_RUN flag
  up@ tcb.status dup @  ~evt-mask and  swap !
;
))
code clr-event-run	\ -- ; reset own EVENT_RUN flag
\ *G Reset the current task's EVENT_RUN flag.
  bic    # evt-mask  0 tcb.status (up)
  ret
end-code

: to-event      \ xt task -- ; define action of a task
\ *G Sets XT as the event handler for the task.
  tcb.event !				\ set into TCB of task
;

[then]


\ ****************
\ *S Message handling
\ ****************
\ *P Message handling is only compiled if the equate
\ ** *\fo{MESSAGE-HANDLER?} is set non-zero in the control
\ ** file.

message-handler? [if]
: msg?		\ task -- flag ; true if task has message
\ *G Returns true if task has received a message.
  tcb.status @ msg-mask and 0<>
;

internal

((
: (send-msg)	\ mesg task -- ; store message in task
  self over tcb.msrc !			\ set source
  dup
  tcb.status dup @  msg-mask or  swap !	\ set message bit
  tcb.mesg !				\ set message
;
))
code (send-msg)	\ mesg task -- ; store message in task
  mov    up  0 tcb.msrc (tos)		\ set source
  mov    @psp+  0 tcb.mesg (tos)	\ set message
  bis    # msg-mask  0 tcb.status (tos)	\ set message bit
  mov    @psp+ tos			\ restore TOS
  ret
end-code

((
: (get-message)	\ -- mesg task
  up@ tcb.mesg @			\ message
  up@ tcb.msrc @
  up@ tcb.status
  dup @  ~msg-mask and  swap !		\ reset message bit
;
))
code (get-message)	\ -- mesg task
  sub    # 4 psp			\ preserve TOS
  mov    tos  2 (psp)
  mov    0 tcb.mesg (up)  0 (psp)	\ get message
  mov    0 tcb.msrc (up)  tos		\ get task
  bic    # msg-mask  0 tcb.status (up)	\ clear message bit
  ret
end-code

external

: send-message  \ addr task -- ; send message to task
\ *G Send a message to a task.
  begin  dup msg?               	\ if task already has a message
    while  pause                	\ wait
  repeat                        	\ and try again
  (send-msg)				\ send the message
;

: get-message   \ -- addr task ; wait for any message
\ *G Wait for any message and return the message and the task it
\ ** came from.
  begin  self msg? 0=           	\ check for a pending message
    while  pause                	\ wait for it
  repeat                        	\ try again
  (get-message)                 	\ read it
;
[then]


event-handler? message-handler? and [if]
: wait-event/msg        \ -- ; wait for message or event trigger
\ *G Wait for a message or an event trigger.
  begin
    self msg?  self event?  or 0=
   while
    pause
  repeat
;
[then]


\ ****************************
\ *S Task structure management
\ ****************************

code init-task	\ xt task -- ; Initialise a task stack
\ *G Initialise a task's stack before running it and
\ ** set it to execute the word whose XT is given.
(( from scheduler
                                        \ save IP - already done with call here
  push   tos				\ save TOS
  push   psp				\ save PSP
  push   lp				\ save LP
  mov    rsp  0 tcb.ssp (up)		\ save SP in TCB
))
  mov    tos  r14			\ r14 := task's RP
  add    # task-r0 task-u0 - #8 -  r14	\ allow for 4 items on return stack

  mov    @psp+  6 (r14)			\ IP := xt
  mov    # 0  4 (r14)			\ TOS
  mov    tos  r13			\ PSP
  add    # task-s0 tos-cached? sp-guard + cells - task-u0 -  r13
  mov    r13  2 (r14)
  mov    r14  0 (r14)			\ LP

  mov    r13  u# s0 (tos)		\ set initial value of S0
  mov    tos  r12			\ r12 := task's RP
  add    # task-r0 task-u0 -  r12	\
  mov    r12  u# r0 (tos)		\ set in R0
  mov    # #10  u# base (tos)		\ BASE := 10

  mov    r14  0 tcb.ssp (tos)		\ preserve RSP in TCB
  mov    # 0  0 tcb.status (tos)	\ clear task status
  mov    # 0  0 tcb.msrc (tos)		\ clear messages
  mov    # 0  0 tcb.mesg (tos)

  mov    @psp+ tos			\ restore TOS
  ret
end-code

: add-task      \ task -- ; insert into list
\ *G Add the task to the list of tasks after the current task.
  self tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  self tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ *G Remove the task from the task list.
\ look for task which points to required task
  self					\ head of chain ; -- target current
  begin  2dup tcb.link @ <>		\ while curr. does not point to target
   while				\ step through chain
    tcb.link @  dup self =		\ but if task not in list, check for end
    if  2drop  exit  endif		\ and get out
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

external

: initiate	\ xt task -- ; start task from scratch
\ *G Start the given task executing the word whose XT is given, e.g.
\ *C   ['] <name> <task> INITIATE
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  run-mask swap tcb.status !		\ mark as running
  pause					\ let it run
;

: terminate	\ task -- ; stop task, and remove from list
\ *G Stop a task, and remove it from the list.
  sub-task pause			\ halt task, and remove from chain
;

: init-multi    \ -- ; initialisation with multi-tasking
\ *G Initialise the multitasker and start it.
\ ** Make sure that your initialisation code includes
\ ** *\fo{INIT-MULTI} or your code will crash.
  single				\ tasker disabled
  main /tcb 0 fill			\ clean TCB
  main main tcb.link !			\ initialise MAIN to point to itself
  run-mask main tcb.status !		\ and reset status field
  multi					\ mark tasker enabled
;

: his		\ task uservar -- addr ; produce address of user var in another task
\ *G Given a task id and a *\fo{USER} variable, returns the
\ ** address of that variable in the given task. This word is
\ ** used to set up *\fo{USER} variables in other tasks.
  self - +
;


\ **********
\ *S Semaphores
\ **********
\ *P The semaphore code is only compiled if the equate
\ ** *\fo{SEMAPHORES?} is set non-zero in the control file.

\ *P A *\fo{SEMAPHORE} is an extended variable used for signalling
\ ** between tasks, and for resource allocation.
\ ** The counter field is used as a count of the number of times
\ ** the resource may be used, and the arbiter field contains the
\ ** *\fo{TCB} of the task that last gained access. This field is
\ ** used for priority arbitration and deadlock detection/arbitration.

semaphores? [if]

[undefined] [I [if]
code [I		\ R: -- x
\ *G Preserve I/F status on return stack, and disable interrupts.
  pop    r14
  push   sr
  dint
  br     r14
end-code

code I]		\ R: x --
\ *G Restore interrupt status from the return stack.
  pop    r14
  pop    sr
  br     r14
end-code
[then]


interpreter
: semaphore     \ -- ; -- addr [child]
  idata create
    0 ,  0 ,				\ count and arbiter fields
  cdata
;
target

heads? [if]
: semaphore	\ -- ; -- addr [child]
\ *G Creates a semaphore which returns its address at runtime.
\ ** Use in the form:
\ *C Semaphore <name>
  create
    0 ,  0 ,				\ count and arbiter fields
;
[then]

: signal        \ addr --
\ *G *\fo{SIGNAL} increments the counter field of a semaphore,
\ ** indicating either that another item has been allocated to
\ ** the resource, or that it is available for use again, 0
\ ** indicating in use by a task
  [I					\ must be interrupt safe
    dup incr  cell+ off			\ inc. counter, release
  I]
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ ** *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a COUNTED semaphore. For
\ ** example a character buffer may be used where the semaphore counter
\ ** shows the number of available characters.
\ ** Alternatively the semaphore may be used purely to share
\ ** resources. The semaphore is initialised to one. The first
\ ** task to *\fo{REQUEST} it gains access, and all other tasks
\ ** must wait until the accessing task *\fo{SIGNAL}s that it has
\ ** finished with the resource.
  begin
    [I  dup @ 0=			\ n.b test and set
   while
    I]  pause				\ operations must be
  repeat                                \ non-interruptible and indivisible
  dup decr				\ got it, decrement counter
  self swap cell+ !			\ mark resource as mine
  I]					\ re-enable interrupts
;

[then]


\ ***************
\ *S TASK and START:
\ **************
\ *P *\fo{TASK <name>} builds a named task user area.
\ ** The action of a task is assigned and the task started
\ ** by the word *\fo{INITIATE}.
\ *C   ['] <action> <task> INITIATE

\ *P *\fo{START:} is used inside a colon definition. The code
\ ** before *\fo{START:} is the task's initialisation, performed
\ ** by the current task. The code after *\fo{START:} up to the
\ ** closing *\fo{;} is the action of the task. For example:
\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin ... pause again
\ **   ;

\ *P All tasks must run in an endless loop, except for initialisation
\ ** code. When *\fo{RUN-FOO} is executed, the code after *\fo{START:}
\ ** is set up as the action of task *\fo{FOO} and started.
\ ** *\fo{RUN-FOO} then exits.

\ *P If you want to perform additional actions after starting the
\ ** task, you* should use *\fo{INITIATE} to start the task.


interpreter	\ host version
: task		\ -- ; -- task ; TASK <name> builds a task
  udata task-size reserve constant	\ build pointer
  cdata
;
target

heads? [if]	\ target version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Creates a new task and data area, returning the address
\ ** of the user area at run time. The task is also linked into
\ ** the task chain anchored by *\fo{TASK-CHAIN}.
  here task-size allot constant		\ build pointer
;
[then]

: start:	\ task -- ; exits from caller
\ *G Used inside a colon definition. The code following
\ ** *\fo{START:} up to the ending semi-colon forms the action
\ ** of the task. The word containing *\fo{START:} finishes at
\ ** *\fo{START:}.
  r> swap initiate
;


\ ********
\ all done
\ ********

decimal


\ *********
\ Test code
\ *********

test-multi? [if]

task task1	\ -- addr ; build task area

#20000 value rate1	\ -- n ; number of idle iterations

: action1	\ -- ; action of task1
  console opvec !
  begin
    [char] * emit
    rate1 0 do pause loop
  again
;

\ ' action1 task1 initiate		\ use this to start the task

task task2	\ -- addr ; build task area

#30000 value rate2	\ -- n ; number of idle iterations

: run-task2	\ -- ; example of using START:
  task2 start:
    console opvec !
    begin
      [char] / emit
      rate2 0 do pause loop
    again
;

[then]

decimal


\ ======
\ *> ###
\ ======

