\ MPE High level ANS kernel for embedded systems

((
Released for use with the MPE Forth Cross Compiler by:
and copyright (c) 1998-2004, 2006, 2008, 2010, 2012, 2013, 2014

MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com


To do
=====
Remove compiler checking and security to reduce code size.

Factor COLD into initialisation and Forth entry.

Remove in-place usage of UPPER because it can cause writes to Flash.
UCMOVE may be useful.

Change NEXT-CASE to be compatible with VFX Forth.

Remove uses of PAD where possible.

Change error codes to ANS codes.


Change history
==============
20140122 SFP016 EMPTY ends with a reboot.
20131219 SFP015 Removed WIDTH. Added ORG.
		Rewrote NUMBER? and friends.
20120825 SFP014 Converted to Lite form.
20101215 SFP013 Made >NUMBER case-insensitive.
20100623 MPE012 Changed POSTPONE to be smaller and less sensitive
		to target architecture.
20090704 MPE011 Improved overlap detection in MOVE so that we can
		make better use of optimised CMOVEs, e.g. ARM.
20081218 SFP010 Removed [IODEV and IODEV].
		Added [IO and IO].
20060307 MPE009 Added compiler macro for >THREADS.
20040401 MPE008 Added NEXTCASE.
20040225 MPE007 Bulletproofed PLACE.
20040218 MPE006 Added SETCONSOLE for 32 bit systems.
20030320 SFP005 Added FLUSHKEYS for 32 bit tagets.
                Modified ACCEPT and HALT? to ignore LF characters.
20020910 SFP004 Removed ERROR and ?ERROR
20020704 SFP003 Corrected REPEAT - how did this survive so long?
20011112 SFP002 Unmarked: start of DOCGEN documentation - an ongoing
                process.
20000614 SFP001 Unmarked reordering to reduce forward refs and size.
))


\ =========
\ *! kernellite
\ *T High level kernel - kernel72lite.fth.
\ =========
\ *P The Forth kernel words documented here are entirely written
\ ** in high-level Forth. The kernel is reduced in size to match
\ ** available code size in small devices such as the MSP430G2553
\ ** in the TI Launchpad.

only forth definitions
decimal

\ -----------------
\ character equates
\ -----------------

$07 equ ABELL                           \ sound
$08 equ BSIN                            \ back space from key
$7F equ DELIN                           \ delete from key
$08 equ BSOUT                           \ back space for emit
$09 equ ATAB                            \ tab
$0D equ ACR                             \ carriage return
$0A equ ALF                             \ line feed
$0C equ FFEED                           \ form feed
$20 equ ABL                             \ space
$2E equ ADOT                            \ .


\ *****************
\ *S User variables
\ *****************

internal
variable next-user	\ -- addr
\ *G Next valid offset for a *\fo{USER} variable created by *\fo{+USER}.
  next-user off
external

interpreter
: +user-offset	\ -- offset ; user offset of next +USER allocation
  next-user @
;

: +user		\ size --
\ *G Used in the cross compiler to create a *\fo{USER} variable
\ ** *\i{size} bytes long at the next available offset and
\ ** updates that offset.
  next-user @ user
  next-user +!
;
target

tcb-size +user SELF		\ task identifier and TCB
\ *G When multitasking is installed,
\ ** the task control block for a task occupies *\fo{TCB-SIZE} bytes at
\ ** the start of the user area. Thus the user area pointer
\ ** also acts as a pointer to the task control block.
+user-offset equ s0-offset
cell +user S0		\ base of data stack
\ *G Holds the initial setting of the data stack pointer.
\ ** N.B. *\fo{S0}, *\fo{R0}, *\fo{#TIB} and *\fo{'TIB} must be
\ ** defined in that order.
+user-offset equ r0-offset
cell +user R0           \ base of return stack
\ *G Holds the initial setting of the return stack pointer.
cell +user #TIB		\ number of chars currently in TIB
\ *G Holds the number of characters currently in *\fo{TIB}.
cell +user 'TIB		\ address of TIB
\ *G Holds the address of *\fo{TIB}, the terminal input buffer.
cell +user >IN		\ offset into TIB
\ *G Holds the current character position being processed in the
\ ** input stream.
internal
cell +user XON/XOFF	\ true if XON/XOFF protocol in use
\ +G True when console is using XON/XOFF protocol.
cell +user ECHOING	\ true if echoing
\ +G True when console is echoing input characters.
external
cell +user OUT		\ number of chars displayed on current line
\ *G Holds the number of chars displayed on current output line.
\ ** Reset by CR.
internal
cell +user BASE		\ current numeric conversion base
\ +G Holds the current numeric conversion base.
\ +* INTERNAL.
cell +user HLD		\ used during number formatting
\ +G Holds data used during number formatting.
\ +* INTERNAL.
cell +user HANDLER	\ used in catch and throw
\ +G Holds the address of the previous exception frame.
cell +user DPL		\ position of double number character id
\ *G Holds the number of characters after the double number indicator
\ ** character. *\fo{DPL} is initialised to -1, which indicates a
\ ** single number, and is incremented for each character after
\ ** the separator.
external
((
cell +user OPVEC	\ output vector
\ *G Holds the address of the I/O vector for the current output device.
cell +user IPVEC	\ input vector
\ *G Holds the address of the I/O vector for the current input device.
))
internal
cell +user 'AbortText   \ Address of text from ABORT"
\ +G Set by the run-time action of *\fo{ABORT"} to hold the address
\ +* of the counted string used by *\fo{ABORT" <text>"}.
external

#34 dup next-user +!  equ PICNUMSIZE
#64 chars dup +user PAD
\ *G A temporary string scratch buffer.
  equ PADSIZE

cr ." Kernel USER area size is " next-user @ .


\ **************
\ *S System data
\ **************

\ ============
\ *N Constants
\ ============

$20 constant BL	\ -- char
\ *G A blank space character.

$40 equ C/L				\ characters/line for display functions

$11 equ XON                             \ XON for flow control
$13 equ XOFF                            \ XOFF for flow control

2 equ #VOCS                             \ maximum number of vocabularies in search order
#VOCS cells equ VSIZE                   \ maximum size of search order area CONTEXT


\ ============
\ *N System variables and data
\ ============
\ *P Note that *\fo{FENCE}, *\fo{DP}, *\fo{RP} and *\fo{VOC-LINK}
\ ** must be declared in that order.

ihere equ DefStart	\ --addr
\ +G Start address of saved Forth state.

internal
\ variable FENCE		\ -- addr
\ +G End of the protected dictionary.
external
variable DP		\ -- addr
\ *G Flash dictionary pointer.
variable RP		\ -- addr
\ *G RAM dictionary pointer.
internal
variable VOC-LINK	\ -- addr
\ +G Links vocabularies.
\ +* INTERNAL.
variable CSP		\ -- addr
\ +G Preserved stack pointer for compile time error checking.
\ +* INTERNAL.
external
variable xDP  DP xDP !	\ -- addr
\ *G Holds the address of the current dictionary pointer,
\ ** *\fo{DP} or *\fo{RP}.
internal
variable STATE		\ -- addr
\ +G Interpreting=0 or compiling=-1.
variable CURRENT	\ -- addr
\ +G Vocabulary/wordlist in which to put new definitions
variable CONTEXT	\ -- addr
\ +G Search order array.
  vsize cell - allot-ram \ search order array
variable LAST		\ -- addr
\ *G Points to name field of last definition
defer start-action	\ --
\ +G The initial action of an application.
  assign noop to-do start-action
vocabulary FORTH
\ +G The standard general purpose vocabulary. INTERNAL.
' forth >body @ equ FORTH-WORDLIST	\ -- wid
FORTH-WORDLIST @ 1+ cells equ /Forth-Wordlist	\ -- len

ihere DefStart - equ /DefStart
\ The initial kernel Forth state is saved in the first INFO sector/page.
\ The application Forth state is saved in the second info sector.

variable PrevName	\ -- addr
\ +G Holds the name field of the previous word in the thread
\ +* between *\fo{HIDE} and *\fo{REVEAL}

external


\ ========================
\ *S Vectored I/O handling
\ ========================

\ ===============
\ *N Introduction
\ ===============
\ *P The standard console Forth I/O words (*\fo{KEY?}, *\fo{KEY},
\ ** *\fo{EMIT}, *\fo{TYPE} and *\fo{CR}) can be used with any I/O
\ ** device by placing the address of a table of xts in the
\ ** *\fo{USER} variables *\fo{IPVEC} and *\fo{OPVEC}. *\fo{IPVEC}
\ ** (input vector) controls the actions of *\fo{KEY?} and *\fo{KEY},
\ ** and *\fo{OPVEC}(output vector) controls the actions of *\fo{EMIT},
\ ** *\fo{TYPE} and *\fo{CR}.
\ ** Adding a new device is matter of writing the five primitives,
\ ** building the table, and storing the address of the table in
\ ** the pointers *\fo{IPVEC} and *\fo{OPVEC} to make the new
\ ** device active. Any initialisation must be performed before
\ ** the device is made active.

\ *P Note that for the output words (*\fo{EMIT}, *\fo{TYPE} and
\ ** *\fo{CR}) the *\fo{USER} variable *\fo{OUT} is handled in the
\ ** kernel before the funtion in the table is called.

\ ==========================
\ *N Building a vector table
\ ==========================
\ *P The example below is taken from an ARM implementation.
\ *E create Console1      \ -- addr
\ **   ' serkey1i ,             \ -- char
\ **   ' serkey?1i ,            \ -- flag
\ **   ' seremit1 ,             \ char --
\ **   ' sertype1 ,             \ c-addr len --
\ **   ' serCR1 ,               \ --
\ **
\ ** Console1 opvec !  Console1 ipvec !

\ ====================
\ *N Generic I/O words
\ ====================

((	\ easier to expand
: ipfunc	\ n -- ; i*x -- j*x
  create
    cells ,
  does>
    @ ipvec @ + @ execute
;

: opfunc	\ n -- ; i*x -- j*x
  create
    cells ,
  does>
    @ opvec @ + @ execute
;

0 ipfunc KEY	\ -- char ; receive char
1 ipfunc KEY?	\ -- flag ; check receive char
2 opfunc EMIT	\ -- char ; display char
3 opfunc TYPE	\ caddr len -- ; display string
4 opfunc CR	\ -- ; display new line
5 opfunc TYPEC	\ caddr len -- ; display string in code space (Harvard only)
))

((
: key	        \ -- char ; receive char
\ *G Wait until the current input device receives a character and
\ ** return it.
  ipvec @ ( 0 cells + ) @ execute  ;
: KEY?		\ -- flag ; check receive char
\ *G Return true if a character is available at the current input
\ ** device.
  ipvec @ cell+  @ execute  ;
: EMIT		\ -- char ; display char
\ *G Display char on the current I/O device. *\fo{OUT} is incremented
\ ** before executing the vector function.
  1 out +!
  opvec @ [ 2 cells ] literal + @ execute  ;
: TYPE		\ caddr len -- ; display string
\ *G Display/write the string on the current output device.
\ ** *\i{Len} is added to *\fo{OUT} before executing the vector
\ ** function.
  dup out +!
  opvec @ [ 3 cells ] literal + @ execute  ;
: CR		\ -- ; display new line
\ *G Perform the equivalent of a CR/LF pair on the current output
\ ** device. *\fo{OUT} is zeroed. before executing the vector
\ ** function.
  out off
  opvec @ [ 4 cells ] literal + @ execute  ;
))

: SPACE         \ --
\ *G Output a blank space (ASCII 32) character.
  bl emit
;

: SPACES        \ n --
\ *G Output *\i{n} spaces, where *\i{n} > 0.
\ ** If *\i{n} < 0, no action is taken.
  0 max 0 ?do  space  loop
;


\ ************************
\ *S Laying data in memory
\ ************************
\ *P These words are used to control and place data in memory.
\ ** Note that the Forth system compiles headers and code into
\ ** Flash memory.

: HERE          \ -- addr
\ *G Return the current dictionary pointer which is the first
\ ** address-unit of free space within the system.
  xdp @ @
;

: ORG		\ addr --
\ *G Set the current dictionary pointer.
  xdp @ !
;

: ALLOT         \ n --
\ *G Allocate N address-units of data space from the current value of
\ ** *\fo{HERE} and move the pointer.
  xdp @ +!
;

: RHERE		\ -- addr
\ *G Return the current RAM dictionary pointer.
  rp @
;

: RALLOT         \ n --
\ *G Allocate n bytes of RAM from *\fo{RHERE} and move the pointer.
  rp +!
;

: ROM		\ --
\ *G *\fo{HERE}, *\fo{ORG}, *\fo{ALLOT}, *\fo{,} and friends, are
\ ** set to use the Flash dictionary pointer. This is the default.
  DP xDP !
;

: RAM		\ --
\ *G *\fo{HERE}, *\fo{ORG} and *\fo{ALLOT} are set to use
\ ** the RAM dictionary pointer. Use in the form:
\ *C   RAM ... ROM
  RP xDP !
;

internal
: [ROM		\ -- x
\ +G Force ROM and save state.
  xDP @  ROM
;

: ROM]		\ x --
\ +G Restore ROM/RAM state.
  xDP !
;

: ROM?		\ -- flag
\ +G True if ROM set.
  xDP @ DP =
;
external

: ALIGN         \ --
\ *G *\fo{ALIGN} dictionary pointer using the same rules as
\ ** *\fo{ALIGNED}.
  here aligned org
;

: ,             \ x --
\ *G Place the CELL value X into the dictionary at *\fo{HERE} and
\ ** increment the pointer.
  here !f  cell allot
;

: C,            \ char --
\ *G Place the CHAR value into the dictionary at *\fo{HERE} and
\ ** increment the pointer.
  here c!f  1 allot
;


\ ************************
\ *S Dictionary management
\ ************************
\ *P The Forth header is laid out as below. The start and end of
\ ** the header are aligned at cell boundaries.
\ *E Link | Count | <name>
\ ** ----------------------
\ ** Cell |  Byte | n Bytes
\ ** ----------------------
\ *D Link       Also called LFA. This field contains the address of the
\ **            of the next count byte in the same thread of the wordlist.
\ *D Count/Ctrl The bottom five bits contain the length (0..31)
\ **            of the name in bytes. The top three bits are used
\ **            as follows:
\ *E            Bit 7    Always set
\ **            Bit 6    Immediate bit (0=immediate)
\ **            Bit 5    Reserved
\ *D <name>     A string of ASCII characters which make up the
\ **            name of the word..

compiler
: N>LINK        \ a-addr -- a-addr'
  cell-
;

: LINK>N        \ a-addr -- a-addr'
  cell+
;

: >#threads ;	\ -- ; compiler treats this as an immediate noop

: >THREADS      \ wid -- a-addr ; MPE009
  >#threads cell+
;

: >vocname
  cell-
;

: >VOC-LINK     \ wid -- a-addr
  2 cells -
;

: latest	( -- caddr ) Last @  ;

target

internal
: LATEST        \ -- c-addr
\ +G Return the address of the name field of the last definition.
  Last @
;
: HIDE		\ --
\ +G Unhook the last word from the current thread and save it
\ +* in *\fo{PrevName}.
  Last @ dup PrevName !			\ save previous state
  n>link @ Last !			\ use previous word
;
: REVEAL	\ --
\ +G Relink the hidden word.
  PrevName @ Last !
;
external

-short-branches
: FIND          \ c-addr -- c-addr 0|xt 1|xt -1
\ *G Perform the *\fo{SEARCH-WORDLIST} operation on all wordlists
\ ** within the current search order. This definition takes a
\ ** counted string rather than a *\i{c-addr/u} pair. The counted
\ ** string is returned as well as the 0 on failure.
  dup count upper                       \ make sure search text is upper case
  dup c@ if                             \ if there is any search text
    context vsize bounds do		\  for all word lists in search order:
      dup count i @ search-wordlist     \    search word list for text
      ?dup if                           \    if text is found then bail out now
        rot drop 			\      discard text address
	unloop  exit
      then
    cell +loop
  then
  0
;
+short-branches

: .NAME         \ nfa --
\ *G Display a definition's name given an NFA.
  count $1F and type space
;

internal
: header,	\ --
\ +G Create a new header in the dictionary. INTERNAL
  bl word count  2dup upper		\ convert to upper case - must be in RAM
\ check for redefinitions
  2dup current @ search-wordlist	\ see if name exists
  if  >name CR .NAME ." is redefined "  then
\ form hash and create link
  align                                 \ force DP to even address
  over c@ over +			\ -- caddr len hash ; first char + count
  current @                             \ -- caddr len hash wid ; get wid = voc-cfa
  dup >#threads @ 1- rot and cells      \ mask for #threads
  swap >threads +                       \ -- caddr len thread ; form thread addr
  dup @ ,                               \ lay old thread ; thread is in RAM
  here swap !				\ -- caddr len ; link this word - here is new NFA
\ lay count + string
  here last !                           \ NFA of latest word
  $1F min dup $C0 or c,			\ count byte, bits 7 and 6 set
  bounds ?do  i c@ c,  loop
  align					\ xt must be even
;

: dataAddr,	\ flag --
\ +G Lay the data address for a defining word where the flag is
\ +* true for compiling to Flash.
  if  dp @ cell +  else  rp @  endif
  ,
;
external

: CREATE        \ --
\ *G Create a new definition in the dictionary. When the new
\ ** definition is executed it will return the address of the
\ ** definition's data area. As compilation is into Flash,
\ ** *\fo{CREATE} cannot be used with *\fo{DOES>} and
\ ** *\fo{ <BUILDS ... DOES> ...} must be used instead.
  ROM? [ROM swap			\ old flag
  header, DOCREATE compile, dataAddr,	\ header, m/c call, address
  ROM]
;

: <BUILDS	\ --
\ *G Always used in the form:
\ *C  : defword <BUILDS ... DOES> ... ;
\ *P When *\fo{defword} is executed a new definition is created
\ ** with the data defined between *\fo{<BUILDS} and *\fo{DOES>}
\ ** and the action defined between *\fo{DOES>} and *\fo{;}.
\ ** You must use *\fo{<BUILDS} and *\fo{DOES>} together, otherwise
\ ** there will be a crash. Treat *\fo{<BUILDS} as a special case
\ ** of *\fo{CREATE} for use with *\fo{DOES>} and compilation into
\ ** Flash.
  ROM? [ROM swap			\ old flag
  header, (dobuild,) dataAddr,		\ header, m/c call, address
  ROM]
;


\ ---------------------
\ *S String compilation
\ ---------------------

internal
: (C")          \ -- c-addr
\ *G The run-time action for *\fo{C"} which returns the address of, and
\ ** steps over, a counted string.
\ ** INTERNAL.
  (")
;

: (S")          \ -- c-addr u
\ *G The run-time action for *\fo{S"} which returns the address and
\ ** length of, and steps over, a string.
\ ** INTERNAL.
  (") count
;

: (ABORT")      \ i*x x1 -- | i*x
\ *G The run time action of *\fo{ABORT"}.
\ ** INTERNAL.
  (") swap if
    'AbortText !  -2 throw
  else
    drop
  then
;

: (.")          \ --
\ *G The run-time action of *\fo{."}.
\ ** INTERNAL.
  (") count type
;
external


\ ============================
\ *S ANS words CATCH and THROW
\ ============================
\ *P *\fo{CATCH} and *\fo{THROW} form the basis of all Forth error
\ ** handling. The following description of *\fo{CATCH} and
\ ** *\fo{THROW} originates with Mitch Bradley and is taken from
\ ** an ANS Forth standard draft.

\ *P *\fo{CATCH} and *\fo{THROW} provide a reliable mechanism for
\ ** handling exceptions, without having to propagate exception
\ ** flags through multiple levels of word nesting. It is similar
\ ** in spirit to the "non-local return" mechanisms of many other
\ ** languages, such as C's *\b{setjmp()} and *\b{longjmp()},
\ ** and LISP's *\b{CATCH} and *\b{THROW}. In the Forth context,
\ ** *\fo{THROW} may be described as a "multi-level EXIT", with
\ ** *\fo{CATCH} marking a location to which a *\fo{THROW} may return.

\ *P Several similar Forth "multi-level EXIT" exception-handling
\ ** schemes have been described and used in past years. It is not
\ ** possible to implement such a scheme using only standard words
\ ** (other than *\fo{CATCH} and *\fo{THROW}), because there is no
\ ** portable way to "unwind" the return stack to a predetermined
\ ** place.

\ *P *\fo{THROW} also provides a convenient implementation technique
\ ** for the standard words *\fo{ABORT} and *\fo{ABORT"}, allowing
\ ** an application to define, through the use of *\fo{CATCH}, the
\ ** behavior in the event of a system abort.

\ ==============
\ *N Example use
\ ==============
\ *P If *\fo{THROW} is executed with a non zero argument, the
\ ** effect is as if the corresponding *\fo{CATCH} had returned it.
\ ** In that case, the stack depth is the same as it was just
\ ** before *\fo{CATCH} began execution.
\ ** The values of the i*x stack arguments could have been modified
\ ** arbitrarily during the execution of xt.  In general, nothing useful
\ ** may be done with those stack items, but since their number is known
\ ** (because the stack depth is deterministic), the application may
\ ** *\fo{DROP} them to return to a predictable stack state.

\ *P Typical use:
\ *E : could-fail	\ -- char
\ **   KEY DUP [CHAR] Q =
\ **   IF  1 THROW  THEN
\ ** ;
\ **
\ ** : do-it		\ a b -- c
\ **   2DROP could-fail
\ ** ;
\ **
\ ** : try-it		\ --
\ **   1 2 ['] do-it CATCH IF
\ **     ( -- x1 x2 ) 2DROP ." There was an exception" CR
\ **   ELSE
\ **     ." The character was " EMIT CR
\ **   THEN
\ ** ;
\ **
\ ** : retry-it		\ --
\ **   BEGIN
\ **     1 2 ['] do-it CATCH
\ **    WHILE
\ **     ( -- x1 x2 ) 2DROP  ." Exception, keep trying" CR
\ **   REPEAT ( char )
\ **   ." The character was " EMIT CR
\ ** ;

\ ==========
\ *N Gotchas
\ ==========
\ *P If a *\fo{THROW} is performed without a *\fo{CATCH} in place,
\ ** the system will/may crash. As the current exception frame
\ ** is pointed to by the *\fo{USER} variable *\fo{HANDLER},
\ ** each task and interrupt handler will need a *\fo{CATCH} if
\ ** *\fo{THROW} is used inside it.

\ *P You can no longer use *\fo{ABORT} as a way of resetting the
\ ** data stack and calling *\fo{QUIT}. *\fo{ABORT} is now defined
\ ** as *\fo{-1 THROW}.

\ =============
\ *N User words
\ =============

: CATCH         \ i*x xt -- j*x 0|i*x n
\ *G Execute the code at XT with an exception frame protecting it.
\ ** *\fo{CATCH} returns a 0 if no error has occurred, otherwise it
\ ** returns the throw-code passed to the last *\fo{THROW}.
  sp@ >r                                \ save PSP
  lp@ >r				\ save LP
  handler @ >r                          \ and last handler
  rp@ handler !                         \ set new handler
  execute                               \ this gets back if there's no THROW
  r> handler !                          \ restore previous handler
  r> drop				\ discard saved LP
  r> drop                               \ discard saved PSP
  0                                     \ good exit = 0
;

: THROW         \ k*x n -- k*x|i*x n
\ *G Throw a non-zero exception code n back to the last *\fo{CATCH}
\ ** call. If n is 0, no action is taken except to *\fo{DROP} n.
  ?dup 0= if  exit  then		\ zero = no throw
  handler @ rp!				\ restore last RSP
  r> handler !				\ restore prev handler
  r> lp!				\ restore LP
  r> swap >r				\ n > RS
  sp! drop r>				\ restore DS
					\ Returns to caller of CATCH
;

: ?throw        \ flag throw-code -- ; SFP017
\ *G Perform a *\fo{THROW} of value throw-code if flag is non-zero,
\ ** otherwise do nothing except discard flag and throw-code.
  swap
  if  throw  else  drop  endif
;

: ABORT"        \ Comp: "ccc<quote>" -- ; Run: i*x x1 -- | i*x ; R: j*x -- | j*x
\ *G If x1 is non-zero at run-time, store the address of the
\ ** following counted string in *\fo{USER} variable *\fo{'ABORTTEXT},
\ ** and perform *\fo{-2 THROW}. The text interpreter in *\fo{QUIT}
\ ** will (if reached) display the text.
  ( ?comp )  compile (abort")  ",
; IMMEDIATE


\ ********************************
\ *S Formatted and unformatted i/o
\ ********************************

\ =======================
\ *N Setting number bases
\ =======================

: HEX           \ --
\ *G Change current radix to base 16.
  #16 base !
;

: DECIMAL       \ --
\ *G Change current radix to base 10.
  #10 base !
;

: BIN		\ --
\ *G Change current radix to base 2.
  #10 base !
;

compiler
: binary	( -- )  2 base !  ;
: bin   	( -- )  2 base !  ;
target

\ =================
\ *N Numeric output
\ =================

: HOLD          \ char --
\ *G Insert the ASCII 'char' value into the pictured numeric output
\ ** string currently being assembled.
  -1 hld +!  hld @ c!
;

((
: SIGN          \ n --
 \ *G Insert the ascii 'minus' symbol into the numeric output string if
 \ ** 'n' is negative.
  0< if  [char] - hold  then
;
))

: #             \ ud1 -- ud2
\ *G Given a double number on the stack this will add the next
\ ** digit to the pictured numeric output buffer and return the
\ ** next double number to work with. *\b{PLEASE NOTE} that the
\ ** numeric output string is built from *\b{right} (l.s. digit)
\ ** to *\b{left} (m.s. digit).
  base @ mu/mod rot 9 over <
  if  7 +  then
  [char] 0 + hold
;

: #S            \ ud1 -- ud2
\ *G Keep performing *\fo{#} until all digits are generated.
  begin  #  2dup or 0=  until
;

: <#            \ --
( *G Begin definition of a new numeric output string buffer.            )
  pad hld !
;

: #>            \ xd -- c-addr u
\ *G Terminate defnition of a numeric output string. Return the
\ ** address and length of the ASCII string.
  2drop  hld @ pad over -  ;

((
: -TRAILING     \ c-addr u1 -- c-addr u2
 \ *G Modify a string address/length pair to ignore any trailing spaces.
  dup 0 ?do
    2dup + 1- c@ bl <> ?leave
    1-
  loop
;
))

: D.R           \ d n --
\ *G Output the double number 'd' using current radix, right justified
\ ** to 'n' characters. Padding is inserted using spaces on the left
\ ** side.
  >r swap over dabs <#
    #s rot 0<
    if  [char] - hold  then
  #>
  r> over - spaces type
;

: D.            \ d --
\ *G Output the double number 'd' without padding.
  0 d.r space
;

: .             \ n --
\ *G Output the cell signed value 'n' without justification.
  s>d d.
;

: U.            \ u --
\ *G As with . but treat as unsigned.
  0 d.
;

((
: U.R           \ u n --
 \ *G As with D.R but uses a single-unsigned cell value.
  0 swap d.r
;
))

: .R            \ n1 n2 --
\ *G As *\fo{D.R} but uses a single-signed cell value.
  >r s>d r> d.r
;


\ ================
\ *N Numeric input
\ ================

internal
: +DIGIT        \ d1 n -- d2 ; accumulates digit into double accumulator
\ *G Multiply d1 by the current radix and add n to it.
\ ** INTERNAL.
  swap base @ um* drop                  \ accumulate into double
  rot base @ um*  d+
  ( 1 #d +! )  dpl @ 1+                     \ incr. #digits, double?
  if  1 dpl +!  then			\ yes, incr. after d.p.
;
external

: >NUMBER       \ ud1 c-addr1 u1 -- ud2 c-addr2 u2 ; convert all until non-digits
\ *G Accumulate digits from string c-addr1/u2 into double number ud1
\ ** to produce ud2 until the first non-convertible character is found.
\ ** c-addr2/u2 represents the remaining string with c-addr2 pointing
\ ** the non-convertible character. The number base for conversion is
\ ** defined by the contents of *\fo{USER} variable *\fo{BASE}.
\ ** *\fo{>NUMBER} is case insensitive.
  dup 0 ?do
    over c@ upc base @ digit 0= ?leave	\ convert char to digit, get out if failed
    -rot 2>r +digit 2r>			\ accumulate digit
    1 /string				\ bump address, decr char count
  loop
;

internal

: (INTEGER?)	\ c-addr u -- d/n/- 2/1/0
\ *G The guts of *\fo{INTEGER?} but without the base override handling.
\ ** See *\fo{INTEGER?}
\ ** INTERNAL.
  -1 dpl !  over c@ [char] - = dup >r
  if  1 /string  endif
  dup 1 < if  2drop 0  r> drop exit  endif
  0 0 2swap begin			\ -- d caddr len
    >number dup
   while
    1 /string  case  over 1- c@
      [char] : of  endof		\ ignore
      [char] . of  dpl off  endof	\ . for double numbers
        drop  2drop 2drop 0  r> drop exit
    endcase
  repeat
  2drop
\  r> ?dnegate
  r> if  dnegate  then
  dpl @ 0<
  if  drop 1  else  2  endif
;

: Check-Prefix  \ addr len -- addr' len'
\ *G If any *\fo{BASE} override prefices or suffices are used
\ ** in the input string, set *\fo{BASE} accordingly and return
\ ** the string without the override characters.
\ ** INTERNAL.
  Case  over c@                         \ get possible prefix char
    [char] $ Of 1 /string  hex     Endof \  $ for HEX
    [char] # Of 1 /string  decimal Endof \  # for DECIMAL
    [char] % Of 1 /string  binary  Endof \  % for BINARY
  EndCase
;

external

: number?	\ $addr -- n 1 | d 2 | 0
\ *G Attempt to convert the counted string at 'addr' to an integer.
\ ** The return result is either 0 for failed, 1 for a single-cell
\ ** return result followed by that cell, or 2 for a double return.
\ ** The ASCII number string supplied can also contain implicit radix
\ ** over-rides. A leading $ enforces hexadecimal, a leading # enforces
\ ** decimal and a leading % enforces binary. Character literals can be
\ ** obtained with 'x' where x is the character.
  count dup if				\ if something to do
    dup 3 = if				\ 'c' check
      over c@ [char] ' = if
        over 2 + c@ [char] ' = if
          drop 1 + c@ 1 exit
        endif
      endif
    endif
    2dup upper				\   convert to upper case
    base @ >r                           \   preserve current base
    check-prefix (integer?)             \   check prefix, convert to number
    r> base !                           \   restore base
  Else
    2drop  ( #l off )  0
  Endif
;


\ ==========================
\ *S String input and output
\ ==========================

internal

: BS            \ -- ; destructive backspace
\ *G Perform a destructive backspace by issuing ASCII characters
\ ** 8, 20h, 8. If *\fo{OUT} is non-zero at the start, it is decremented
\ ** by one regardless of the actions of the device driver.
\ ** INTERNAL.
  out @ ?dup if
    bsout emit  space  bsout emit
    1- out !
  then
;

: ?BS           \ pos -- pos' step ; perform BS if pos non-zero
\ *G If pos is non-zero and *\fo{ECHOING} is set, perform *\fo{BS}
\ ** and return the size of the step, 0 or -1.
\ ** INTERNAL.
  dup 0<>  echoing @ 0<>  and if        \ pos non-zero and echo
    bs 1- -1                            \   perform backspace
  else
    0                                   \   no
  then
;

: SAVE-CH       \ char addr -- ; save as required
\ *G Save char at addr, and output the character if
\ ** *\fo{ECHOING} is set.
\ ** INTERNAL.
  over swap c!				\ stash character
  echoing @                             \ echo it?
  if  emit  else  drop  then            \ process accordingly
;

external

: ."            \ "ccc<quote>" --
\ *G Output the text upto the closing double-quotes character.
\ ** Use *\fo{.( <text>)} when interpreting.
  ['] (.") compile,  ",
;
IMMEDIATE

((
: $.            \ c-addr -- ; display counted string
\ *G Output a counted-string to the output device.
  count type
;
))

: ACCEPT	\ c-addr +n1 -- +n2 ; read up to LEN chars into ADDR
\ *G Read a string of maximum size n1 characters to the buffer at
\ ** c-addr, returning n2 the number of characters actually read. Input
\ ** may be terminated by CR. The action may be input device specific.
\ ** If *\fo{ECHOING} is non-zero, characters are echoed.
\ ** If *\fo{XON/XOFF} is non-zero, an XON character is sent at the
\ ** start and an XOFF character is sent at the the end.
  dup 0=				\ check for pathological case
  if  nip exit  then
  xon/xoff @ if  xon emit  then         \ enable server
  0 -rot bounds                         \ form address limits
  ?do
    case key                            \ get character
      bsin  of ?bs               endof  \ BS
      atab  of bl i save-ch 1+ 1 endof  \ TAB, convert to space
\      alf   of bl i save-ch 1+ 1 endof  \ LF, convert to space
      alf   of 0 endof			\ LF, ignore
      acr   of bl i save-ch  leave endof  \ CR
      ffeed of bl i save-ch  leave endof  \ FF, treat like CR
      delin of ?bs               endof  \ DEL
             i save-ch  1+ 1 0		\ stash char, echo it?
    endcase
  +loop                                 \ round again
  xon/xoff @ if  xoff emit  then        \ disable server
;


\ ***********************
\ *S Source input control
\ ***********************

internal
variable <source>

((
: SOURCE-ID	\ -- n ; indicates input source
\ *G Returns an indicator of which device is generating
\ ** source input. See the ANS specification for more details.
  <source> @
;
))

compiler
: SOURCE-ID	( -- n )  <source> @  ;
: TO-SOURCE	( c-addr u -- )  #tib !  'tib !  ;
: SOURCE        ( -- c-addr u )  'tib @  #tib @  ;
target

((
internal
: SOURCE        \ -- c-addr u
\ *G Returns the address and length of the current terminal input
\ ** buffer.
\ ** INTERNAL
  'tib @  #tib @
;
external
))

: QUERY         \ -- ; fetch line into TIB
( *G Reset the input source specification to the console and accept a   )
( ** line of text into the input buffer.                                )
  0 <source> !                          \ mark source as keyboard
  'tib @ tib-len 1- accept #tib !	\ "tib-len 1-" was #80
  >in off
;


\ ================
\ *S Text scanning
\ ================

: PARSE         \ char "ccc<char>" -- c-addr u
\ *G Parse the next token from the terminal input buffer using
\ ** <char> as the delimiter. The next token is returned as a
\ ** *\fo{c-addr/u} string description. Note that *\fo{PARSE} does
\ ** not skip leading delimiters. If you need to skip leading
\ ** delimiters, use *\fo{PARSE-WORD} instead.
  >r
  source >in @ /string                  \ remainder of input is substring
  over swap r> scan                     \ look for char
  >r over -                             \ subtract addresses to get length
  dup r> 0<> -                          \ if string found then bump >IN past char
  >in +!                                \ update >IN to reflect parsed text
;

internal
: PARSE-WORD    \ char -- c-addr u ; find token, skip leading chars
\ *G An alternative to *\fo{WORD} below. The return is a
\ ** *\i{c-addr/u} pair rather than a counted string and no copy
\ ** has occured, i.e. the contents of *\fo{HERE} are unaffected.
\ ** Because no intermediate global buffers are used *\fo{PARSE-WORD}
\ ** is more reliable than *\fo{WORD} for text scanning in multi-threaded
\ ** applications.
\ ** INTERNAL.
  >r  source tuck  >in @ /string
  r@ skip  over swap r> scan
  >r over -  rot r>  dup 0<> + - >in !
;
external

: WORD          \ char "<chars>ccc<char>" -- c-addr
\ *G Similar behaviour to *\fo{PARSE-WORD} but the returned
\ ** string is a counted string.
  parse-word  rp @ place  rp @
;


\ ----------------
\ *S Miscellaneous
\ ----------------

internal

: HALT?         \ -- flag
 \ *G Used in listed displays. This word will check the keyboard for a
 \ ** 'pause' key <space>, if the key is pressed it will then wait for
 \ ** a continue key or an abort key. The return flag is TRUE if abort
 \ ** is requested. Line Feed (LF, ASCII 10) characters are ignored.
  key? dup 0=				\ key not pressed?
  if  exit  endif
  key bl = if				\ space?
    drop				\   chuck previous flag
    key bl <>				\   get second char
  then
;

external

((
internal

origin [if]	\ only if ORIGIN is non-zero
: origin-	\ addr -- addr'
\ +G If addr is non-zero, subtract the start address of the first
\ +* defined CDATA section.
\ +* This word is only compiled if the start address of the first
\ +* defined CDATA section is non-zero.
\ +* INTERNAL.
  dup if  origin -  then                \ normalize if non-zero
;

compiler
: origin+	\ addr -- addr' ; denormalise NFA again
\ +G If addr is non-zero, add the start address of the first
\ +* defined CDATA section.
\ +* This word is only compiled if the start address of the first
\ +* defined CDATA section is non-zero.
\ +* INTERNAL.
  dup if  origin +  then		\ normalize if non-zero
;
target
[else]
compiler	\ treat these as immediate noops
: origin-  ;
: origin+  ;
target
[then]

: nfa-buff      \ -- addr+len addr ; make a buffer for holding NFAs
\ +G Form a temporary buffer for holding NFAs. A factor for
\ +* *\fo{WORDS}.
\ +* INTERNAL.
  pad                                   \ buffer
  context @ >#threads @ cells           \ length
  bounds
;

: MAX-NFA       \ -- addr c-addr ; returns addr and top nfa
\ +G Return the thread address and NFA of the highest word
\ +* in the NFA buffer. A factor for *\fo{WORDS}.
\ +* INTERNAL.
  0 0                                   \ clear result
  nfa-buff do
    i @  over u> if			\ if greater
      2drop  i  i @                     \ replace current set
    then
  cell +loop                            \ step on
  origin+				\ de-normalize
;

: COPY-THREADS  \ addr --
\ +G Copy the threads of the *\fo{CONTEXT} wordlist to a
\ +* temporary NFA buffer for manipulation. A factor for
\ +* *\fo{WORDS}.
\ +* INTERNAL.
  context @ >threads                    \ source
  nfa-buff do                           \ destination/len
    dup @ origin- i !
    cell+
  cell +loop                            \ copy and normalise
  drop                                  \ discard source addr
;

external

: WORDS         \ --
\ *G Display the names of all definitions in the wordlist at the
\ ** top of the search-order.
  cr                                    \ set up
  copy-threads                          \ copy out & normalize
  begin
    max-nfa dup                         \ while still one left
    halt? 0= and                        \ and no stop command
   while
    out @ c/l > if  cr  then            \ new line if needed
    dup .name                           \ show name
    n>link @ origin- swap !		\ find next
  repeat
  2drop
;
))

: WORDS         \ --
\ *G Display the names of all the words in the dictionary. The
\ ** order of the words may not be as expected as we are trying
\ ** to save code space in Flash.
  context @ dup >threads
  swap >#threads @ cells		\ addr len
  bounds ?do
    i @ begin				\ -- nfa
      dup
     while
      out @ c/l > if  cr  then		\ new line if needed
      dup .name				\ show name
      n>link @				\ next name
    repeat
    drop
  cell +loop
;

: MOVE          \ addr1 addr2 u -- ; intelligent move
\ *G An intelligent memory move, chooses between *\fo{CMOVE} and
\ ** *\fo{CMOVE>} at runtime to avoid memory overlap problems.
\ ** Note that as ROM PowerForth characters are 8 bit, there is
\ ** an implicit connection between a byte and a character.
  >r  2dup swap - r@ u< if		\ overlap if dest-src<len ; MPE011
    r> cmove>				\ move down
  else
    r> cmove				\ move up
  then
;

: DEPTH         \ ??? -- +n
\ *G Return the number of items on the data stack.
  s0 @ sp@ - cell - cell /
;

: .FREE         \ --
\ *G Return the free dictionary space.
  cr APPFLASHEND dp @ - u. ." bytes free Flash"
  cr RP-END rp @ -      u. ." bytes free RAM"
;


\ -------------------
\ *S Wordlist control
\ -------------------

asmcode
here is-action-of vocabulary	\ --
\ *G The runtime action of a *\fo{VOCABULARY}.
  pop  r14              	\ get address of address
  mov  @r14 r14			\ get address
  mov  @r14 & context		\ set contents
  ret
end-code


\ ---------------------
\ *S Control structures
\ ---------------------

internal
: !CSP          \ x --
\ *G Save the stack pointer in *\fo{CSP}.
\ ** Used for on-target compile-time error checking.
\ ** INTERNAL.
  sp@ csp !
;

: ?CSP          \ --
\ *G Issue an error if the stack pointer is not the
\ ** same as the value previously stored in CSP.
\ ** Used for on-target compile-time error checking.
\ ** INTERNAL.
  sp@ csp @ <> #-22 ?throw
;

((
: ?PAIRS        \ x1 x2 --
\ *G If x1<>x2, issue an error.
\ ** Used for on-target compile-time error checking.
\ ** INTERNAL.
  <> #-22 ?throw
;
))

((
: ?COMP         \ --
\ *G Error if not in compile state.
\ ** INTERNAL.
  state @ 0= #-14 ?throw
;
))

((
: ?EXEC         \ --
\ *G Error if not interpreting.
\ ** INTERNAL.
  state @ #-403 ?throw
;
))
external

: DO            \ C: -- do-sys ; Run: n1|u1 n2|u2 -- ; R: -- loop-sys
\ *G Begin a *\fo{DO ... LOOP} construct. Takes the end-value and
\ ** start-value from the data-stack.
  ( ?comp )  c_do  ( 3 )
;  IMMEDIATE

: ?DO           \ C: -- do-sys ; Run: n1|u1 n2|u2 -- ; R: -- | loop-sys
\ *G Compile a *\fo{DO} which will only begin loop execution if the
\ ** loop parameters are not the same. Thus *\fo{0 0 ?DO ... LOOP}
\ ** will not execute the contents of the loop.
  ( ?comp )  c_?do  ( 3 )
;  IMMEDIATE

: LOOP          \ C: do-sys -- ; Run: -- ; R: loop-sys1 -- | loop-sys2
\ *G The closing statement of a *\fo{DO..LOOP} construct.
\ ** Increments the index and terminates when the index crosses
\ ** the limit.
  ( ?comp )  ( 3 ?pairs )  c_loop
;  IMMEDIATE

: +LOOP         \ C: do-sys -- ; Run: n -- ; R: loop-sys1 -- | loop-sys2
\ *G As with *\fo{LOOP} except that you specify the increment on
\ ** the data-stack.
  ( ?comp )  ( 3 ?pairs )  c_+loop
;  IMMEDIATE

: BEGIN         \ C: -- dest ; Run: --
\ *G Mark the start of a structure of the form:
\ *C   BEGIN ... [WHILE] ... UNTIL / AGAIN / [REPEAT]
  ( ?comp )  c_mrk_branch<
;  IMMEDIATE

((
: AGAIN         \ C: dest -- ; Run: --
\ *G The end of a *\fo{BEGIN ... AGAIN} construct which specifies
\ ** an infinite loop.                                                              )
  ( ?comp )  c_branch<
;  IMMEDIATE
))

: UNTIL         \ C: dest -- ; Run: x --
\ *G Compile code into definition which will jump back to the
\ ** matching *\fo{BEGIN} if the supplied condition flag is
\ ** Zero/FALSE.
  ( ?comp )  c_?branch<
;  IMMEDIATE

: WHILE         \ C: dest -- orig dest ; Run: x --
\ *G Separate the condition test from the loop code in a
\ ** *\fo{BEGIN ... WHILE ... REPEAT} block.
  ( ?comp )  c_?branch> swap
;  IMMEDIATE

: REPEAT        \ C: orig dest -- ; Run: --
\ *G Loop back to the conditional dest code in a
\ ** *\fo{BEGIN ... WHILE ... REPEAT} construct.                                                         )
  ( ?comp )  c_branch<   >c_res_branch	\ SFP003
;  IMMEDIATE

: IF            \ C: -- orig ; Run: x --
\ *G Mark the start of an *\fo{IF ... [ELSE] ... THEN} conditional
\ ** block.
  ( ?comp )  c_?branch>
;  IMMEDIATE

: THEN          \ C: orig -- ; Run: --
\ *G Mark the end of an *\fo{IF..THEN} or *\fo{IF..ELSE..THEN}
\ ** conditional construct.
  ( ?comp )  >c_res_branch
;  IMMEDIATE

: ELSE          \ C: orig1 -- orig2 ; Run: --
\ *G Begin the failure condition code for an *\fo{IF}.
  ( ?comp )  c_branch>  swap >c_res_branch
;  IMMEDIATE

((
: RECURSE       \ Comp: --
\ *G Compile a recursive call to the colon definition containing
\ ** *\fo{RECURSE} itself. Do not use *\fo{RECURSE} between
\ ** *\fo{DOES>} and *\fo{;}. Used in the form:
\ *C : foo  ... recurse ... ;
\ *P to compile a reference to *\fo{FOO} from inside *\fo{FOO}.
  PrevName @ name> compile,
;  IMMEDIATE
))

\ **********************************
\ *S Target interpreter and compiler
\ **********************************

internal
: ?STACK        \ --
\ *G Error if stack pointer out of range.
\ ** INTERNAL.
  sp@ s0 @ u> #-4 ?throw
;

: ?UNDEF        \ x --
\ *G Word not defined error if x=0.
\ ** INTERNAL.
  0= #-13 ?throw
;
external

: POSTPONE      \ Comp: "<spaces>name" --
\ *G Compile a reference to another word. *\fo{POSTPONE} can handle
\ ** compilation of *\fo{IMMEDIATE} words which would otherwise be
\ ** executed during compilation.
  bl word find dup ?undef		\ it is a name
  -1 = if				\ if non-immediate
    [compile] literal  ['] compile,	\ lay "lit xt  compile,"
  then
  compile,				\ immediate word will be compiled now
; IMMEDIATE

: S"            \ Comp: "ccc<quote>" -- ; Run: -- c-addr u
\ *G Describe a string. Text is taken up to the next double-quote
\ ** character. The address and length of the string are returned.
  state @ if
    compile (s") ",
  else
   [char] " parse  ( pad place  pad count )
  then
; IMMEDIATE

: C"            \ Comp: "ccc<quote>" -- ; Run: -- c-addr
\ *G As *\fo{S"} except the address of a counted string is returned.
  state @ if
    compile (c") ",
  else
    [char] " parse  pad place  pad
  then
;
IMMEDIATE

: LITERAL       \ Comp: x -- ; Run: -- x
\ *G Compile a literal into the current definition. Usually used
\ ** in the form *\fo{[ <expression ] LITERAL} inside a colon
\ ** definition. Note that *\fo{LITERAL} is *\fo{IMMEDIATE}.
  c_lit
;
IMMEDIATE

((
: CHAR          \ "<spaces>name" -- char
( *G Return the first character of the next token in the input stream.	)
( ** Usually used to avoid magic numbers in the source code.		)
  bl word 1+ c@
;

: [CHAR]        \ Comp: "<spaces>name" -- ; Run: -- char
( *G Compile the first character of the next token in the input stream	)
( ** as a literal. Usually used to avoid magic numbers in the source 	)
( ** code.								)
  ( ?comp )  char [compile] literal
;
IMMEDIATE
))

: [             \ --
\ *G Switch compiler into interpreter state.
  state off
;
IMMEDIATE

: ]             \ --
\ *G Switch compiler into compilation state.
  state on
;

: IMMEDIATE     \ --
\ *G Mark the last defined word as *\fo{IMMEDIATE}. Immediate
\ ** words will execute whenever encountered regardless of
\ ** *\fo{STATE}.
  $40 last @ bbic!			\ bit 6 := 0
;

: '             \ "<spaces>name" -- xt
\ *G Find the xt of the next word in the input stream. An error occurs
\ ** if the xt cannot be found.
  bl word find ?undef
;

: [']           \ Comp: "<spaces>name" -- ; Run: -- xt
\ *G Find the xt of the next word in the input stream, and compile it
\ ** as a literal. An error occurs if the xt cannot be found.
  ' [compile] literal
;
IMMEDIATE

((
: [COMPILE]     \ "<spaces>name" --
\ *G Compile the next word in the input stream. *\fo{[COMPILE]}
\ ** ignores the *\fo{IMMEDIATE} state of the word. *\fo{[COMPILE]}
\ ** is mostly superceded by *\fo{POSTPONE}.
  ' compile,
;
IMMEDIATE
))

: (             \ "ccc<paren>" --
\ *G Begin an inline comment. All text upto the closing bracket is
\ ** ignored.
  [char] ) word drop
;
IMMEDIATE

: \             \ "ccc<eol>" --
\ *G Begin a single-line comment. All text up to the end of the line is
\ ** ignored.
  0 word drop				\   use variable length line
;
IMMEDIATE

: ",            \ "ccc<quote>" --
\ *G Parse text up to the closing quote and compile into the
\ ** dictionary at *\fo{HERE} as a counted string. The end of the
\ ** string is aligned.
  [char] " word count dup c,		\ count byte
  bounds ?do  i c@ c,  loop
  align
;

internal
: (TO-DO)       \ -- ; R: xt -- a-addr'
\ *G The run-time action of *\fo{IS}. It is followed by the
\ ** data addres of the *\fo{DEFER}red word at which the xt is stored.
\ ** INTERNAL.
  r> dup cell+ aligned >r
  @ !
;
external

: IS         \ "<spaces>name" --
\ *G The second part of the *\fo{ASSIGN xxx TO-DO yyy} construct.
\ ** This word will assign the given XT to be the action of a
\ ** *\fo{DEFER}ed word which is named in the input stream.
  ' >body  state @ if
    compile (to-do) ,
  else
    !
  then
;
IMMEDIATE

: exit		\ R: nest-sys -- ; exit current definition
\ *G Compile code into the current definition to cause a definition to
\ ** terminate. This is the Forth equivalent to inserting an RTS/RET
\ ** instruction in the middle of an assembler subroutine.
  ( ?comp )  c_exit
; immediate

: ;             \ C: colon-sys -- ; Run: -- ; R: nest-sys --
\ *G Complete the definition of a new 'colon' word or *\fo{:NONAME}
\ ** code block.
  ( ?comp ) ?csp  c_exit
  reveal [compile] [
;
IMMEDIATE

-short-branches
: INTERPRET     \ --
\ *G Process the current input line as if it is text entered at
\ ** the keyboard.
  begin
    ?stack  bl word dup c@
  while
    find ?dup if
      0< state @ and
      if  compile,  else  execute  then
    else
      number? ?dup ?undef
      state @ if
        case
          1 of  c_lit  endof
          2 of  swap c_lit c_lit  endof
            #-13 throw
        endcase
      else
        drop
      then
    then
  repeat
  drop
;
+short-branches

((
: EVALUATE      \ i*x c-addr u -- j*x ; interpret the string
\ *G Process the supplied string as though it had been entered via the
\ ** interpreter.
  'tib @ >r  #tib @ >r  >in @ >  source-id >r
  to-source  -1 <source> !  >in off
  ['] interpret catch
  r> <source> !  r> >in !  r> #tib !  r> 'tib !
  throw
;
))

internal
: .throw	\ throw# --
\ *G Display the throw code. Values of 0 and -1 are ignored.
  case
     0 of  endof				\ quiet
    -1 of  endof				\ quiet
    -2 of  'AbortText @ count type  endof	\ ABORT" ..."
    -13 of  ." is undefined"  endof
       ." Throw code " .
  end-case
;
external

: QUIT		\ -- ; R: i*x --
\ *G Empty the return stack, store 0 in *\fo{SOURCE-ID}, and enter
\ ** interpretation state. *\fo{QUIT} repeatedly *\fo{ACCEPT}s a
\ ** line of input and *\fo{INTERPRET}s it, with a prompt if
\ ** interpreting and *\fo{ECHOING} is on. Note that any task that
\ ** uses *\fo{QUIT} must initialise *\fo{'TIB}, *\fo{BASE},
\ ** *\fo{IPVEC}, and *\fo{OPVEC}.
  xon/xoff off  echoing on              \ No Xon/Xoff, do Echo
  0 <source> !  [compile] [		\ set up
  begin
    r0 @ rp!                            \   reset return stack
    echoing @ if  cr  then              \   if echoing enabled issue new line
    query				\   get user input
    ['] interpret catch ?dup 0= if	\   interpret line
      state @ 0=  echoing @ 0<>  and if	\   if interpreting & echoing
        ."  ok"  depth ?dup		\     prompt user
        if  ." -" .  then
      then
    else
      .throw  s0 @ sp!  [compile] [	\    display error, clean up
      cr source type			\    display input line
      cr >in @ 1- spaces ." ^"		\    display pointer to error
    then
  again                                 \   do next line
;


\ ***************
\ *S Startup code
\ ***************

\ ====================
\ *N The COLD sequence
\ ====================
\ *P At power up, the target executes *\fo{COLD} or the word specified
\ ** by *\fo{MAKE-TURNKEY <name>}, or the word specified as the action
\ ** of an application compiled by the target.

internal

3 cells equ #idhead

: (INIT)        \ --
\ *G Performs the high level Forth startup. See the source code for
\ ** more details.
\ ** INTERNAL.

\ Copy the RAM table from Flash into RAM. The table may
\ initialise multiple blocks of RAM. The data consists of:
\ len, addr, pageid, len bytes data
  init-ram
  dup #idhead +  swap 2@ cmove          \ source, dest, len
((
\ Copy the RAM table from Flash into RAM. The table may
\ initialise multiple blocks of RAM. The data consists of:
\ len, addr, pageid, len bytes data, len, addr, page, ..., len=0,
  init-ram begin
    dup @
   while                                \ len<>0
    dup #idhead +  over 2@              \ source, dest, len
    cmove                               \ copy
    dup @ +  #idhead +                  \ step to next block
  repeat
  drop
))
  user-reset s0 4 cells cmove           \ initialise user vars
  s0 @ sp!                              \ Reset data stack
  init-tib 'tib !			\ input buffer pointer
  handler off				\ no CATCH used yet
  init-dp dp 3 cells cmove		\ dp, rp, voc-link
\  init-fence fence 4 cells cmove        \ fence, dp, rp, voc-link
  0 <source> !  decimal
  forth-wordlist dup context ! current !
  0 context cell+ !
\  only forth definitions                \ set up search order
;
external

: Commit	\ xt|0 --
\ *G Preserve the compiled image. If xt is non-zero, that word
\ ** will be executed when the application starts.
  dup 0= if  drop ['] noop  then
  to-do start-action			\ start up action
  DefStart APPSTART /DefStart >Info	\ save application data
;

: Empty		\ --
\ *G Wipe the application and perform a cold restart.
\  INFOSTART DefStart /DefStart cmove	\ relink base system
  APPSTART /DefStart FlErase		\ wipe application links
  APPFLASHSTART /APPFLASH FlErase	\ wipe application flash
  reboot
;

: COLD          \ --
\ *G The first high level word executed by default. This word is
\ ** set to be the word executed at power up, but this may be
\ ** overridden by a later use of *\fo{MAKE-TURNKEY <name>} in
\ ** the cross-compiled code. See
\ ** the source code for more details of *\fo{COLD}.
  (init)                                \ start Forth
  init-ser                              \ initialise serial line
\  console opvec !			\ default i/o channels
\  console ipvec !
  start-clock
\ perform the application detection and linking
  DefStart INFOSTART /DefStart >Info	\ save defaults
  APPSTART @ dup -1 <> swap 0<> and if  \ FENCE not $FFFF or 0
    crcslot @ APPSTART /DefStart + @ = if   \ checksums match
      APPSTART DefStart /DefStart cmove	    \ load app
    endif
  endif
  dp @ @ -1 <>				\ Flash at here should be erased
  if  empty  endif			\ use -1 for 16/32 bit reasons
\ sign on
  CR .cpu .free				\ sign on, display free space
  start-action				\ perform application start up actions
  cr cr ."   ok"			\ display prompt
  s0 @ sp!                              \ reset data stack
  quit                                  \ start text interpreter
;
make-turnkey cold                       \ Default start-up word.


\ =====================
\ *S Kernel error codes
\ =====================

\ *D -1 ABORT
\ *D -2 ABORT"
\ *D -4 Stack underflow
\ *D -13  Undefined word.
\ *D -14  Attempt to interpret a compile only definition.
\ *D -22  Control structure mismatch - unbalanced control structure.
\ *D -121 Attempt to remove with MARKER or FORGET below FENCE in protected dictionary.
\ *D -403 Attempt to compile an interpret only definition.
\ *D -501 Error if not LOADing from a block.


\ ******
\ *> ###
\ ******

